﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Data;
using C1.Win.C1FlexGrid;

namespace WPFGridsBenchmark
{
    public enum DataSourceType
    {
        ListCollectionView,
        DataView
    }
    public abstract class TestGrid : UserControl
    {
        public static Dictionary<string, Type> Grids;
        public static List<Type> GridsForAutoSizeTesting;
        static TestGrid()
        {
            Grids = new Dictionary<string, Type>();
            GridsForAutoSizeTesting = new List<Type>();
            Grids.Add("Microsoft's DataGrid for WPF", typeof(MSDataGrid));
            GridsForAutoSizeTesting.Add(typeof(MSDataGrid));
            Grids.Add("FlexGrid for WPF", typeof(C1FlexGrid));
            Grids.Add("C1DataGrid for WPF", typeof(C1DataGrid));
            GridsForAutoSizeTesting.Add(typeof(C1DataGrid));
            Grids.Add("FlexGrid for WinForms", typeof(WinFormsFlexGrid));
        }

        protected void SetGrid(object grid)
        {
            this.Content = grid;
        }
        public virtual DataSourceType AcceptableDataSource
        {
            get
            {
                return DataSourceType.ListCollectionView;
            }
        }
        public abstract void Load(IEnumerable data);
        public virtual void Dispose()
        {
            this.Content = null;
        }
        public abstract void Sort(bool ascending);
        public virtual string GetGridInfo()
        {
            if (this.Content != null)
            {
                System.Reflection.AssemblyName asName = this.Content.GetType().Assembly.GetName();
                return asName.Name + "." + asName.Version;
            }
            return String.Empty;
        }
        public abstract void ScrollInView(int rowNumber);

        public abstract Task Filter(bool isActive);
        public abstract Task Group();
    }

    public class MSDataGrid : TestGrid
    {
        DataGrid _grid = null;
        public MSDataGrid(bool fixedColumnWidth)
        {
            _grid = new DataGrid();
            // default column sizing is SizeToHeader, finally sets column width by content
            if (fixedColumnWidth)
            {
                _grid.ColumnWidth = new DataGridLength(150);
            }
            SetGrid(_grid);
        }

        public override void Dispose()
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                cv.Filter = null;
            }
            _grid.ItemsSource = null;
            base.Dispose();
        }

        public override void Load(IEnumerable data)
        {
            _grid.ItemsSource = data;
        }

        public override void Sort(bool ascending)
        {
            ICollectionView dataView = _grid.ItemsSource as ICollectionView;
            using (dataView.DeferRefresh())
            {
                dataView.SortDescriptions.Clear();
                dataView.SortDescriptions.Add(new SortDescription("ID", ascending ? ListSortDirection.Ascending : ListSortDirection.Descending));
            }
        }

        public override void ScrollInView(int rowNumber)
        {
            object item = ((ListCollectionView)_grid.ItemsSource).GetItemAt(rowNumber);
            _grid.ScrollIntoView(item);
        }
        public async override Task Filter(bool isActive)
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                if (isActive)
                    cv.Filter = IsActiveFilter;
                else
                    cv.Filter = IsInactiveFilter;
            }
        }

        bool IsActiveFilter(object customer)
        {
            return ((Customer)customer).Active;
        }

        bool IsInactiveFilter(object customer)
        {
            return !((Customer)customer).Active;
        }
        public override async Task Group()
        {
            ICollectionView cv = _grid.ItemsSource as ICollectionView;
            using (cv.DeferRefresh())
            {
                cv.GroupDescriptions.Clear();
                cv.GroupDescriptions.Add(new PropertyGroupDescription("CountryID"));
            }
        }
    }

    public class C1FlexGrid : TestGrid
    {
        C1.WPF.FlexGrid.C1FlexGrid _grid = null;
        public C1FlexGrid(bool fixedColumnWidth)
        {
            _grid = new C1.WPF.FlexGrid.C1FlexGrid();
            // uses some predefined column width depending on underlying data
            // auto-width is only possible with method call
            _grid.AllowAddNew = true;
            SetGrid(_grid);
        }
 
        public override void Dispose()
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                cv.Filter = null;
            }
            _grid.ItemsSource = null;
            base.Dispose();
        }
        public override void Load(IEnumerable data)
        {
            _grid.ItemsSource = data;
        }

        public override void Sort(bool ascending)
        {
            ICollectionView dataView = _grid.ItemsSource as ICollectionView;
            using (dataView.DeferRefresh())
            {
                dataView.SortDescriptions.Clear();
                dataView.SortDescriptions.Add(new SortDescription("ID", ascending ? ListSortDirection.Ascending : ListSortDirection.Descending));
            }
        }
        public override void ScrollInView(int rowNumber)
        {
            _grid.ScrollIntoView(rowNumber, 0);
        }
        public async override Task Filter(bool isActive)
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                if (isActive)
                    cv.Filter = IsActiveFilter;
                else
                    cv.Filter = IsInactiveFilter;
            }
        }

        bool IsActiveFilter(object customer)
        {
            return ((Customer)customer).Active;
        }

        bool IsInactiveFilter(object customer)
        {
            return !((Customer)customer).Active;
        }

        public override async Task Group()
        {
            ICollectionView cv = _grid.ItemsSource as ICollectionView;
            using (cv.DeferRefresh())
            {
                cv.GroupDescriptions.Clear();
                cv.GroupDescriptions.Add(new PropertyGroupDescription("CountryID"));
            }
        }
    }

    public class C1DataGrid : TestGrid
    {
        C1.WPF.DataGrid.C1DataGrid _grid = null;
        public C1DataGrid(bool fixedColumnWidth)
        {
            _grid = new C1.WPF.DataGrid.C1DataGrid();
            // default column sizing is AutoStar, always shows full content, might add some space depending on screen size and default column sizing
            if (fixedColumnWidth)
            {
                _grid.ColumnWidth = new C1.WPF.DataGrid.DataGridLength(150);
            }
            SetGrid(_grid);
        }
        public override void Dispose()
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                cv.Filter = null;
            }
            _grid.ItemsSource = null;
            base.Dispose();
        }

        public override void Load(IEnumerable data)
        {
            _grid.ItemsSource = data;
        }

        public override void Sort(bool ascending)
        {
            ICollectionView dataView = _grid.ItemsSource as ICollectionView;
            using (dataView.DeferRefresh())
            {
                dataView.SortDescriptions.Clear();
                dataView.SortDescriptions.Add(new SortDescription("ID", ascending ? ListSortDirection.Ascending : ListSortDirection.Descending));
            }
        }
        public override void ScrollInView(int rowNumber)
        {
            _grid.ScrollIntoView(rowNumber, 0);
        }
        public async override Task Filter(bool isActive)
        {
            var cv = _grid.ItemsSource as ListCollectionView;
            if (cv != null)
            {
                if (isActive)
                    cv.Filter = IsActiveFilter;
                else
                    cv.Filter = IsInactiveFilter;
            }
        }

        bool IsActiveFilter(object customer)
        {
            return ((Customer)customer).Active;
        }

        bool IsInactiveFilter(object customer)
        {
            return !((Customer)customer).Active;
        }
        public override async Task Group()
        {
            ICollectionView cv = _grid.ItemsSource as ICollectionView;
            using (cv.DeferRefresh())
            {
                cv.GroupDescriptions.Clear();
                cv.GroupDescriptions.Add(new PropertyGroupDescription("CountryID"));
            }
        }
    }

    public class WinFormsFlexGrid : TestGrid
    {
        C1.Win.C1FlexGrid.C1FlexGrid _grid = null;
        System.Windows.Forms.Integration.WindowsFormsHost _host = null;
        public WinFormsFlexGrid(bool fixedColumnWidth)
        {
            _grid = new C1.Win.C1FlexGrid.C1FlexGrid();
            _host = new System.Windows.Forms.Integration.WindowsFormsHost();
            _host.Background = new System.Windows.Media.SolidColorBrush(System.Windows.Media.Colors.White); // to keep performance high, always use non-transparent background
            _host.Child = _grid;
            // uses some predefined column width depending on underlying data
            // auto-width is only possible with method call
            _grid.AllowAddNew = true;
            SetGrid(_host);
        }

        public override void Dispose()
        {
            var cv = _grid.DataSource as DataView;
            cv.RowFilter = null;
            _grid.DataSource = null;
            _host.Dispose();
            _host = null;
            base.Dispose();
        }

        public override DataSourceType AcceptableDataSource
        {
            get
            {
                return DataSourceType.DataView;
            }
        }

        public override void Load(IEnumerable data)
        {
            _grid.DataSource = data;
        }

        public override void Sort(bool ascending)
        {
            _grid.Sort(ascending ? C1.Win.C1FlexGrid.SortFlags.Ascending : C1.Win.C1FlexGrid.SortFlags.Descending, 1);
        }
        public override void ScrollInView(int rowNumber)
        {
            _grid.ShowCell(rowNumber, _grid.Cols.Fixed);
        }
        public async override Task Filter(bool isActive)
        {
            var cv = _grid.DataSource as DataView;
            cv.RowFilter = ("Active = " + isActive.ToString(System.Globalization.CultureInfo.InvariantCulture));
        }

        public override async Task Group()
        {
            _grid.BeginUpdate();
            _grid.Sort(SortFlags.Ascending, 4);
            _grid.Tree.Column = 0;
            _grid.Subtotal(AggregateEnum.None, 0, 4, 4);
            _grid.EndUpdate();
            _grid.Tree.Show(1);
        }

        public override string GetGridInfo()
        {
            if (this._grid != null)
            {
                System.Reflection.AssemblyName asName = _grid.GetType().Assembly.GetName();
                return asName.Name + "." + asName.Version;
            }
            return String.Empty;
        }

    }

}
